import os
os.environ["CUDA_VISIBLE_DEVICES"] = "0"

# tensorflow 1.15.0
import tensorflow
print(tensorflow.__version__)

import tensorflow as tf
from tensorflow.keras import Input
from tensorflow.keras.layers import Conv2D, Flatten, Dense, Conv2DTranspose, Reshape, Lambda, Activation, BatchNormalization, LeakyReLU, Dropout, ZeroPadding2D, UpSampling2D, Embedding, multiply
from tensorflow.keras.models import Model, Sequential
from tensorflow.keras import backend as K
from tensorflow.keras.optimizers import Adam, RMSprop
from tensorflow.keras.callbacks import ModelCheckpoint
from tensorflow.keras.initializers import RandomNormal
from tensorflow.keras.utils import to_categorical
from functools import partial

# Commented out IPython magic to ensure Python compatibility.
import numpy as np
import pandas as pd
import os
import matplotlib
import matplotlib.pyplot as plt
# %matplotlib inline

from datetime import datetime, timezone, timedelta
tz = timezone(timedelta(hours=+8))
now = datetime.now(tz)
str_now = now.strftime("%b_%d_%Y_%H_%M_%S")
print(str_now)

number_of_fold = 2
str_now = "fold_" + str(number_of_fold) + "_" + str_now
str_now

# run params
SECTION = 'gan'
RUN_FOLDER = 'run/{}/'.format(SECTION)
RUN_FOLDER = os.path.join(RUN_FOLDER, str_now)
print(RUN_FOLDER)


if not os.path.exists(RUN_FOLDER):
    os.mkdir(RUN_FOLDER)
    os.mkdir(os.path.join(RUN_FOLDER, 'viz'))
    os.mkdir(os.path.join(RUN_FOLDER, 'images'))
    os.mkdir(os.path.join(RUN_FOLDER, 'weights'))

"""## data"""

BATCH_SIZE = 61
IMAGE_SIZE_H = 80
IMAGE_SIZE_W = 128

from tensorflow.keras.preprocessing.image import ImageDataGenerator
# data_name = DATA_NAME
batch_size = BATCH_SIZE

train_dirName_of_number_of_fold = "fold_" + str(number_of_fold) + "_train"
test_dirName_of_number_of_fold = "fold_" + str(number_of_fold) + "_test"
train_data_folder = os.path.join(r"/nfs/python_code/", train_dirName_of_number_of_fold)
val_data_folder = os.path.join(r"/nfs/python_code/", test_dirName_of_number_of_fold)
print("train_data_folder :", train_data_folder)

data_gen_train = ImageDataGenerator(
        rotation_range=10,
        width_shift_range=0.05,
        height_shift_range=0.05,
        zoom_range=0.05,
        horizontal_flip=True,
        fill_mode='nearest',
        preprocessing_function=lambda x: (x.astype('float32') - 127.5) / 127.5 )

iterator = data_gen_train.flow_from_directory(train_data_folder
                                        , color_mode='grayscale' # gray image
                                        , classes = ['c1', 'c2', 'c3', 'c4']
                                        , target_size = (IMAGE_SIZE_H,IMAGE_SIZE_W)
                                        , batch_size = batch_size
                                        , shuffle = True
                                        , class_mode = 'categorical'
                                        , subset = "training")

print("val_data_folder :", val_data_folder)# validation_iterator

data_gen_val = ImageDataGenerator(preprocessing_function=lambda x: (x.astype('float32') - 127.5) / 127.5 )

validation_iterator = data_gen_val.flow_from_directory(val_data_folder
                                        , color_mode='grayscale' # gray image
                                        , classes = ['c1', 'c2', 'c3', 'c4']
                                        , target_size = (IMAGE_SIZE_H,IMAGE_SIZE_W)
                                        , batch_size = batch_size
                                        , shuffle = True
                                        , class_mode = 'categorical'
                                        , subset = "training")

input_dim = (IMAGE_SIZE_H,IMAGE_SIZE_W,1)
critic_conv_filters = [64,128,256,512]
critic_conv_kernel_size = [3,3,3,3]
critic_conv_strides = [2,2,2,2]
critic_batch_norm_momentum = None
critic_activation = 'leaky_relu'
critic_dropout_rate = None
critic_learning_rate = 0.0001

generator_initial_dense_layer_size = (5, 8, 512)
generator_upsample = [1,1,1,1]
generator_conv_filters = [256,128,64,1]
generator_conv_kernel_size = [3,3,3,3]
generator_conv_strides = [2,2,2,2]
generator_batch_norm_momentum = 0.9
generator_activation = 'leaky_relu'
generator_dropout_rate = None
generator_learning_rate = 0.0001
optimizer = 'adam'
z_dim = 100
batch_size = BATCH_SIZE



n_layers_critic = len(critic_conv_filters)
n_layers_generator = len(generator_conv_filters)

weight_init = RandomNormal(mean=0., stddev=0.02)


d_losses = []
g_losses = []

generator = []
critic = []
num_classes = 4

epoch_counter = 0

epoch_list = []
train_acc_list = []
val_acc_list = []

class RandomWeightedAverage( tensorflow.keras.layers.Layer ):
    def __init__(self, batch_size):
        super().__init__()
        self.batch_size = batch_size

    def _merge_function(self, inputs):
        alpha = K.random_uniform((self.batch_size, 1, 1, 1))
        return (alpha * inputs[0]) + ((1 - alpha) * inputs[1])


def gradient_penalty_loss(y_true, y_pred, interpolated_samples):
    # compute {1 - b[ (grad1)^2 + (grad2)^2 ] }^2
    
    # gradient
    g = K.gradients(y_pred, interpolated_samples)
    gradients = g[0]
    
    # euclidean norm
    gradients_sqr = K.square(gradients)

    # sum
    gradients_sqr_sum = K.sum(gradients_sqr, axis=np.arange(1, len(gradients_sqr.shape)))

    # sqrt
    gradient_l2_norm = K.sqrt(gradients_sqr_sum)

    # lambda * (1 - ||grad||)^2
    gradient_penalty = K.square(1 - gradient_l2_norm)

    # loss over all batch samples
    gradient_penalty_mean =  K.mean(gradient_penalty)

    return gradient_penalty_mean

def wasserstein(y_true, y_pred):
    return -K.mean(y_true * y_pred)

def get_activation(activation):
    if activation == 'leaky_relu':
        layer = LeakyReLU(alpha = 0.2)
    else:
        layer = Activation(activation)
    return layer

def _build_critic():
    critic_input = Input(shape=input_dim, name='input_dim__critic_input')
    x = critic_input

    for i in range(n_layers_critic):
        x = Conv2D(
            filters = critic_conv_filters[i]
            , kernel_size = critic_conv_kernel_size[i]
            , strides = critic_conv_strides[i]
            , padding = 'same'
            , name = 'critic_conv_' + str(i)
            , kernel_initializer = weight_init
            )(x)
        if critic_batch_norm_momentum and i > 0:
            x = BatchNormalization(momentum = critic_batch_norm_momentum)(x)
        x = get_activation(critic_activation)(x)
        if critic_dropout_rate:
            x = Dropout(rate = critic_dropout_rate)(x)
    x = Flatten()(x)
    fake = Dense(1, activation=None
    , kernel_initializer = weight_init
    )(x)
    
    aux = Dense(num_classes, activation='softmax', name='auxiliary', kernel_initializer = weight_init)(x)
    critic = Model(critic_input, [fake, aux], name='critic')
    critic.summary()
    return critic

def _build_generator():
    latent = Input(shape=(z_dim,), name='z_dim__generator_input')
    image_class = Input(shape=(1,), dtype='int32', name='image_class__generator_input')
    
    classes = Embedding(num_classes, z_dim,
                    embeddings_initializer='glorot_normal')(image_class)
    h = multiply([latent, classes])
    x = h
    x = Dense(np.prod(generator_initial_dense_layer_size), kernel_initializer = weight_init)(x)
    if generator_batch_norm_momentum:
        x = BatchNormalization(momentum = generator_batch_norm_momentum)(x)
    x = get_activation(generator_activation)(x)
    x = Reshape(generator_initial_dense_layer_size)(x)

    if generator_dropout_rate:
        x = Dropout(rate = generator_dropout_rate)(x)

    for i in range(n_layers_generator):
        if generator_upsample[i] == 2:
            x = UpSampling2D()(x)
            x = Conv2D(
            filters = generator_conv_filters[i]
            , kernel_size = generator_conv_kernel_size[i]
            , padding = 'same'
            , name = 'generator_conv_' + str(i)
            , kernel_initializer = weight_init
            )(x)
        else:
            x = Conv2DTranspose(
                filters = generator_conv_filters[i]
                , kernel_size = generator_conv_kernel_size[i]
                , padding = 'same'
                , strides = generator_conv_strides[i]
                , name = 'generator_conv_' + str(i)
                , kernel_initializer = weight_init
                )(x)

        if i < n_layers_generator - 1:
            if generator_batch_norm_momentum:
                x = BatchNormalization(momentum = generator_batch_norm_momentum)(x)
            x = get_activation(generator_activation)(x)
        else:
            x = Activation('tanh')(x)

    generator_output = x
    generator = Model([latent, image_class], generator_output, name='generator')
    generator.summary()
    return generator


def get_opti(lr):
    if optimizer == 'adam':
        opti = Adam(lr=lr, beta_1=0.5)
    elif optimizer == 'rmsprop':
        opti = RMSprop(lr=lr)
    else:
        opti = Adam(lr=lr)

    return opti


def set_trainable(m, val):
    m.trainable = val
    for l in m.layers:
        l.trainable = val

def train_critic(true_imgs, true_labels, batch_size, using_generator):
    valid = np.ones((batch_size,1), dtype=np.float32)
    fake = -np.ones((batch_size,1), dtype=np.float32)
    dummy = np.zeros((batch_size, 1), dtype=np.float32)
    dont_care = np.random.randint(0, num_classes, batch_size)
    noise = np.random.normal(0, 1, (batch_size, z_dim))
    sampled_labels = np.random.randint(0, num_classes, batch_size)
    d_loss = critic_model.train_on_batch([true_imgs, noise, sampled_labels], 
                                         [valid, true_labels, fake, to_categorical(sampled_labels), dummy, to_categorical(dont_care)])
    return d_loss

def train_generator(batch_size):
    valid = np.ones((batch_size,1), dtype=np.float32)
    noise = np.random.normal(0, 1, (batch_size, z_dim))
    sampled_labels = np.random.randint(0, num_classes, batch_size)
    return model.train_on_batch([noise, sampled_labels], [valid, to_categorical(sampled_labels)])

def test_critic_performance(true_imgs, true_labels, batch_size):
    valid = np.ones((batch_size,1), dtype=np.float32)
    fake = -np.ones((batch_size,1), dtype=np.float32)
    dummy = np.zeros((batch_size, 1), dtype=np.float32)
    dont_care = np.random.randint(0, num_classes, batch_size)
    noise = np.random.normal(0, 1, (batch_size, z_dim))
    sampled_labels = np.random.randint(0, num_classes, batch_size)
    
    test_d_loss = critic_model.test_on_batch([true_imgs, noise, sampled_labels], 
                                             [valid, true_labels, fake, to_categorical(sampled_labels), dummy, to_categorical(dont_care)])
    
    return test_d_loss

def critic_predict(imgs, onehot_labels):
    labels = np.argmax(onehot_labels, axis=1)
    predict_labels = np.argmax(critic.predict_on_batch(imgs)[1], axis=1)
    correct_count = np.sum(labels == predict_labels)
    return correct_count, len(labels), predict_labels



def train(iterator, batch_size, epochs, run_folder, print_every_n_batches = 250
, n_critic = 3
, using_generator = False):
    global epoch_counter

    for epoch in range(epoch_counter, epoch_counter + epochs):
        if epoch % 100 == 0:
            critic_loops = 5
        else:
            critic_loops = n_critic

        for _ in range(critic_loops):
            if using_generator:
                true_imgs, true_labels = iterator.next()
                if true_imgs.shape[0] != batch_size:
                    true_imgs, true_labels = iterator.next()
            d_loss = train_critic(true_imgs, true_labels, batch_size, using_generator)
        g_loss = train_generator(batch_size)
        d_losses.append(d_loss)
        g_losses.append(g_loss)

        val_imgs, val_labels = validation_iterator.next()
        if val_imgs.shape[0] != batch_size:
            val_imgs, val_labels = validation_iterator.next()
        test_d_loss = test_critic_performance(val_imgs, val_labels, batch_size)
        
        train_correct_count, train_len, predict_labels_trueImgs = critic_predict(true_imgs, true_labels)
        val_correct_count, val_len, predict_labels_valImgs = critic_predict(val_imgs, val_labels)
        val_labels_1D = np.argmax(val_labels, axis=1)
        
        print("epoch", epoch, ", %d/%d train_acc = %.3f, %d/%d val_acc = %.3f" % (train_correct_count, train_len, train_correct_count/val_len, 
                                                                                        val_correct_count, val_len, val_correct_count/val_len) )

        epoch_list.append(epoch_counter)
        train_acc_list.append(train_correct_count/val_len)
        val_acc_list.append(val_correct_count/val_len)
        dataframe = pd.DataFrame({'epoch':epoch_list, 'train_accuracy':train_acc_list, 'val_accuracy':val_acc_list, 'd_loss':d_losses, 'g_loss':g_losses})
        dataframe = dataframe.round({'train_accuracy': 3, 'val_accuracy': 3})
        dataframe.to_csv(RUN_FOLDER + "/train_val_acc.csv")

        
        
        # If at save interval => save generated image samples
        if epoch==10 or epoch % print_every_n_batches == 0 or (train_acc_list[epoch] >= 0.6 and val_acc_list[epoch] >= 0.7):
            print("val_labels_1D:", val_labels_1D)
            print("predict_labels_valImgs:", predict_labels_valImgs)
            predict_dataframe = pd.DataFrame({'val_labels_1D':val_labels_1D, 'predict_labels_valImgs':predict_labels_valImgs})
            predict_dataframe.to_csv(os.path.join(run_folder, 'weights/predict_labels-%d.csv' % (epoch)))
            
            print("Model Save")
            sample_images(true_imgs, true_labels, run_folder)
            model.save_weights(os.path.join(run_folder, 'weights/weights-%d.h5' % (epoch)))
            critic.save_weights(os.path.join(run_folder, 'weights/critic-%d.h5' % (epoch)))
            generator.save_weights(os.path.join(run_folder, 'weights/generator-%d.h5' % (epoch)))

        epoch_counter+=1


def sample_images(true_imgs, true_labels, run_folder):
    r, c = 5, 4
    noise = np.random.normal(0, 1, (r * c, z_dim))
    elements = [0,1,2,3]
    labels = []
    for i in range(5):
        labels = labels + elements
    gen_imgs = generator.predict([noise, np.array(labels)])
    gen_imgs = 0.5 * (gen_imgs + 1)
    gen_imgs = np.clip(gen_imgs, 0, 1)

    fig, axs = plt.subplots(r, c, figsize=(15,15))
    cnt = 0

    for i in range(r):
        for j in range(c):
            axs[i,j].imshow(np.squeeze(gen_imgs[cnt, :,:,:]), cmap = 'gray_r')
            axs[i,j].axis('off')
            cnt += 1
    fig.savefig(os.path.join(run_folder, "images/sample_%d.png" % epoch_counter))
    plt.show()
    plt.close()

def load_weights(filepath):
    model.load_weights(filepath)
    
def one_hot2int(a):
    result = np.array([np.where(r==1)[0][0] for r in a])
    return result

critic = _build_critic()

generator = _build_generator()

set_trainable(generator, False)

real_img = Input(shape=input_dim, name='real_img__critic_model__input')

z_disc = Input(shape=(z_dim,), name='z_dim__critic_model__input')
image_class = Input(shape=(1,), dtype='int32', name='image_class__critic_model__input')
fake_img = generator([z_disc, image_class]) # generate fake img

fake = critic(fake_img) # predict real or fake
valid = critic(real_img) # predict real or fake

interpolated_img = RandomWeightedAverage(batch_size)([real_img, fake_img])
validity_interpolated = critic(interpolated_img)

partial_gp_loss = partial(gradient_penalty_loss,
                  interpolated_samples=interpolated_img)
partial_gp_loss.__name__ = 'gradient_penalty' # Keras requires function names

critic_model = Model(inputs=[real_img, z_disc, image_class],
                    outputs=[valid, fake, validity_interpolated], # [1, -1, 0]
                    name='critic_model')

critic_model.compile(
    loss=[wasserstein, 'categorical_crossentropy',wasserstein, 'categorical_crossentropy', partial_gp_loss, 'categorical_crossentropy'] # [1, -1, 0]
    ,optimizer=get_opti(critic_learning_rate)
    ,loss_weights=[1, 0.01, 1, 0.01, 10, 0]
    )

critic_model.summary()

set_trainable(critic, False)
set_trainable(generator, True)

z_disc = Input(shape=(z_dim,), name='z_dim__model__input')
image_class = Input(shape=(1,), dtype='int32', name='image_class__critic_model__input')
img = generator([z_disc, image_class])
fake, aux = critic(img)
model = Model(inputs=[z_disc, image_class], outputs=[fake, aux], name='model')

model.compile(optimizer=get_opti(generator_learning_rate)
, loss=[wasserstein, "categorical_crossentropy"]
,loss_weights=[10, 0.01]
)

set_trainable(critic, True)

model.summary()

EPOCHS = 30000
PRINT_EVERY_N_BATCHES = 250
N_CRITIC = 3

for i in range(len(critic_model.loss)):
    print(i, critic_model.loss[i])
print()
for i in range(len(critic_model.output)):
    print(i, critic_model.output[i])

train(     
    iterator
    , batch_size = BATCH_SIZE
    , epochs = EPOCHS
    , run_folder = RUN_FOLDER
    , print_every_n_batches = PRINT_EVERY_N_BATCHES
    , n_critic = N_CRITIC
    , using_generator = True
)